// Copyright Contributors to the DNF5 project.
// Copyright Contributors to the libdnf project.
// SPDX-License-Identifier: GPL-2.0-or-later
//
// This file is part of libdnf: https://github.com/rpm-software-management/libdnf/
//
// Libdnf is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 2 of the License, or
// (at your option) any later version.
//
// Libdnf is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with libdnf.  If not, see <https://www.gnu.org/licenses/>.

#include "CoprFunctionsTest.hpp"

#include "helpers.hpp"

using namespace std;

void CoprFunctionsTest::test_repo_fallbacks() {
    CPPUNIT_ASSERT(vector<string>{"epel-8"} == dnf5::repo_fallbacks("epel-8"));
    CPPUNIT_ASSERT((vector<string>{"almalinux-10", "epel-10"} == dnf5::repo_fallbacks("almalinux-10")));
    CPPUNIT_ASSERT((vector<string>{"rhel-10.4", "rhel-10", "epel-10"} == dnf5::repo_fallbacks("rhel-10.4")));
}

CPPUNIT_TEST_SUITE_REGISTRATION(CoprFunctionsTest);
