// Copyright Contributors to the DNF5 project
// SPDX-License-Identifier: LGPL-2.1-or-later

#ifndef LIBDNF5_REPO_REPO_SACK_PRIVATE_HPP
#define LIBDNF5_REPO_REPO_SACK_PRIVATE_HPP

namespace libdnf5::repo {

constexpr const char * SYSTEM_REPO_NAME = "@System";

}

#endif
